"""
Internal utility for converting arguments from a launch spec or call to wandb launch
into a runnable wandb launch script
"""

import enum
import logging
import os
from shlex import quote
import tempfile
from typing import Any, Dict, List, Optional, Tuple

import wandb
from wandb.apis.internal import Api
from wandb.errors import Error as ExecutionError, LaunchError
from wandb.sdk.lib.runid import generate_id

from . import utils


_logger = logging.getLogger(__name__)

DEFAULT_CONFIG_PATH = "launch_override_config.json"


class LaunchSource(enum.IntEnum):
    WANDB: int = 1
    GIT: int = 2
    LOCAL: int = 3


class LaunchProject(object):
    """A launch project specification."""

    def __init__(
        self,
        uri: str,
        target_entity: str,
        target_project: str,
        name: Optional[str],
        docker_config: Dict[str, Any],
        git_info: Dict[str, str],
        overrides: Dict[str, Any],
    ):

        self.uri = uri
        self.target_entity = target_entity
        self.target_project = target_project
        self.name = name
        self.build_image: bool = docker_config.get("build_image", False)
        self.python_version: Optional[str] = docker_config.get("python_version")
        self._base_image: Optional[str] = docker_config.get("base_image")
        self.docker_image: Optional[str] = docker_config.get("docker_image")
        self.docker_user_id: int = docker_config.get("user_id", 1000)
        self.git_version: Optional[str] = git_info.get("version")
        self.git_repo: Optional[str] = git_info.get("repo")
        self.override_args: Dict[str, Any] = overrides.get("args", {})
        self.override_config: Dict[str, Any] = overrides.get("run_config", {})
        self._runtime: Optional[str] = None
        self.run_id = generate_id()
        self._entry_points: Dict[
            str, EntryPoint
        ] = {}  # todo: keep multiple entrypoint support?
        if "entry_point" in overrides:
            self.add_entry_point(overrides["entry_point"])

        if utils._is_wandb_uri(self.uri):
            self.source = LaunchSource.WANDB
            self.project_dir = tempfile.mkdtemp()
        elif utils._is_git_uri(self.uri):
            self.source = LaunchSource.GIT
            self.project_dir = tempfile.mkdtemp()
        else:
            # assume local
            if not os.path.exists(self.uri):
                raise LaunchError(
                    "Assumed URI supplied is a local path but path is not valid"
                )
            self.source = LaunchSource.LOCAL
            self.project_dir = self.uri

        self.aux_dir = tempfile.mkdtemp()

        self.clear_parameter_run_config_collisions()

    @property
    def base_image(self) -> str:
        """Returns {PROJECT}_base:{PYTHON_VERSION}"""
        # TODO: this should likely be source_project when we have it...
        generated_name = "{}_base:{}".format(
            self.target_project, self.python_version or "3"
        )
        return self._base_image or generated_name

    @property
    def image_name(self) -> str:
        """Returns {PROJECT}_launch the ultimate version will
        be tagged with a sha of the git repo"""
        # TODO: this should likely be source_project when we have it...
        return "{}_launch".format(self.target_project)

    def clear_parameter_run_config_collisions(self) -> None:
        """Clear values from the overide run config values if a matching key exists in the override arguments."""
        if not self.override_config:
            return
        keys = [key for key in self.override_config.keys()]
        for key in keys:
            if self.override_args.get(key):
                del self.override_config[key]

    def get_single_entry_point(self) -> "EntryPoint":
        """Returns the first entrypoint for the project."""
        # assuming project only has 1 entry point, pull that out
        # tmp fn until we figure out if we wanna support multiple entry points or not
        if len(self._entry_points) != 1:
            raise Exception("LaunchProject must have exactly one entry point")
        return list(self._entry_points.values())[0]

    def add_entry_point(self, entry_point: str) -> "EntryPoint":
        """Adds an entry point to the project."""
        _, file_extension = os.path.splitext(entry_point)
        ext_to_cmd = {".py": "python", ".sh": os.environ.get("SHELL", "bash")}
        if file_extension in ext_to_cmd:
            command = "%s %s" % (ext_to_cmd[file_extension], quote(entry_point))
            new_entrypoint = EntryPoint(name=entry_point, command=command)
            self._entry_points[entry_point] = new_entrypoint
            return new_entrypoint
        raise ExecutionError(
            "Could not find {0} among entry points {1} or interpret {0} as a "
            "runnable script. Supported script file extensions: "
            "{2}".format(
                entry_point, list(self._entry_points.keys()), list(ext_to_cmd.keys())
            )
        )

    def get_entry_point(self, entry_point: str) -> "EntryPoint":
        """Gets the entrypoint if its set, or adds it and returns the entrypoint."""
        if entry_point in self._entry_points:
            return self._entry_points[entry_point]
        return self.add_entry_point(entry_point)

    def _fetch_project_local(self, api: Api) -> None:
        """Fetch a project into a local directory, returning the path to the local project directory."""
        assert self.source != LaunchSource.LOCAL
        parsed_uri = self.uri
        if utils._is_wandb_uri(self.uri):
            run_info = utils.fetch_wandb_project_run_info(self.uri, api)
            if not run_info["git"]:
                raise ExecutionError("Run must have git repo associated")
            utils._fetch_git_repo(
                self.project_dir, run_info["git"]["remote"], run_info["git"]["commit"]
            )
            patch = utils.fetch_project_diff(self.uri, api)

            if patch:
                utils.apply_patch(patch, self.project_dir)
            entry_point = run_info.get("codePath", run_info["program"])
            # For cases where the entry point wasn't checked into git
            if not os.path.exists(os.path.join(self.project_dir, entry_point)):
                downloaded_entrypoint = utils.download_entry_point(
                    self.uri, api, entry_point, self.project_dir
                )
                if not downloaded_entrypoint:
                    raise LaunchError(
                        f"Entrypoint: {entry_point} does not exist, "
                        "and could not be downloaded. Please specify the entrypoint for this run."
                    )
                # if the entrypoint is downloaded and inserted into the project dir
                # need to rebuild image with new code
                self.build_image = True

            if entry_point.endswith("ipynb"):
                entry_point = utils.convert_jupyter_notebook_to_script(
                    entry_point, self.project_dir
                )

            # Download any frozen requirements
            utils.download_wandb_python_deps(self.uri, api, self.project_dir)
            # Specify the python runtime for jupyter2docker
            self.python_version = run_info.get("python", "3")

            if not self._entry_points:
                self.add_entry_point(entry_point)

            self.override_args = utils.merge_parameters(
                self.override_args, run_info["args"]
            )
        else:
            assert utils._GIT_URI_REGEX.match(parsed_uri), (
                "Non-wandb URI %s should be a Git URI" % parsed_uri
            )

            if not self._entry_points:
                wandb.termlog(
                    "Entry point for repo not specified, defaulting to main.py"
                )
                self.add_entry_point("main.py")

            utils._fetch_git_repo(self.project_dir, parsed_uri, self.git_version)


class EntryPoint(object):
    """An entry point into a wandb launch specification."""

    def __init__(self, name: str, command: str):
        self.name = name
        self.command = command
        self.parameters: Dict[str, Any] = {}

    def _validate_parameters(self, user_parameters: Dict[str, Any]) -> None:
        missing_params = []
        for name in self.parameters:
            if name not in user_parameters and self.parameters[name].default is None:
                missing_params.append(name)
        if missing_params:
            raise ExecutionError(
                "No value given for missing parameters: %s"
                % ", ".join(["'%s'" % name for name in missing_params])
            )

    def compute_parameters(
        self, user_parameters: Optional[Dict[str, Any]]
    ) -> Tuple[Dict[str, Optional[str]], Dict[str, Optional[str]]]:
        """Validates and sanitizes parameters dict into expected dict format.

        Given a dict mapping user-specified param names to values, computes parameters to
        substitute into the command for this entry point. Returns a tuple (params, extra_params)
        where `params` contains key-value pairs for parameters specified in the entry point
        definition, and `extra_params` contains key-value pairs for additional parameters passed
        by the user.
        """
        if user_parameters is None:
            user_parameters = {}
        # Validate params before attempting to resolve parameter values
        self._validate_parameters(user_parameters)
        final_params = {}
        extra_params = {}

        parameter_keys = list(self.parameters.keys())
        for key in parameter_keys:
            param_obj = self.parameters[key]
            key_position = parameter_keys.index(key)
            value = (
                user_parameters[key]
                if key in user_parameters
                else self.parameters[key].default
            )
            final_params[key] = param_obj.compute_value(value, key_position)
        for key in user_parameters:
            if key not in final_params:
                extra_params[key] = user_parameters[key]
        return (
            self._sanitize_param_dict(final_params),
            self._sanitize_param_dict(extra_params),
        )

    def compute_command(self, user_parameters: Optional[Dict[str, Any]]) -> str:
        """Converts user parameter dictionary to a string."""
        params, extra_params = self.compute_parameters(user_parameters)
        command_with_params = self.command.format(**params)
        command_arr = [command_with_params]
        command_arr.extend(
            [
                "--%s %s" % (key, value) if value is not None else "--%s" % (key)
                for key, value in extra_params.items()
            ]
        )
        return " ".join(command_arr)

    @staticmethod
    def _sanitize_param_dict(param_dict: Dict[str, Any]) -> Dict[str, Optional[str]]:
        """Sanitizes a dictionary of paramaeters, quoting values, except for keys with None values."""
        return {
            (str(key)): (quote(str(value)) if value is not None else None)
            for key, value in param_dict.items()
        }


def get_entry_point_command(
    entry_point: "EntryPoint", parameters: Dict[str, Any]
) -> List[str]:
    """Returns the shell command to execute in order to run the specified entry point.

    Arguments:
    entry_point: Entry point to run
    parameters: Parameters (dictionary) for the entry point command

    Returns:
        List of strings representing the shell command to be executed
    """
    commands = []
    commands.append(entry_point.compute_command(parameters))
    return commands


def create_project_from_spec(launch_spec: Dict[str, Any], api: Api) -> LaunchProject:
    """Constructs a LaunchProject instance using a launch spec.

    Arguments:
    launch_spec: Dictionary representation of launch spec
    api: Instance of wandb.apis.internal Api

    Returns:
        An initialized `LaunchProject` object
    """
    uri = launch_spec["uri"]

    name: Optional[str] = None
    if launch_spec.get("name"):
        name = launch_spec["name"]

    return LaunchProject(
        uri,
        launch_spec["entity"],
        launch_spec["project"],
        name,
        launch_spec.get("docker", {}),
        launch_spec.get("git", {}),
        launch_spec.get("overrides", {}),
    )


def fetch_and_validate_project(
    launch_project: LaunchProject, api: Api
) -> LaunchProject:
    """Fetches a project into a local directory, adds the config values to the directory, and validates the first entrypoint for the project.

    Arguments:
    launch_project: LaunchProject to fetch and validate.
    api: Instance of wandb.apis.internal Api

    Returns:
        A validated `LaunchProject` object.

    """
    if launch_project.source == LaunchSource.LOCAL:
        if not launch_project._entry_points:
            wandb.termlog("Entry point for repo not specified, defaulting to main.py")
            launch_project.add_entry_point("main.py")
    else:
        launch_project._fetch_project_local(api=api)
    first_entry_point = list(launch_project._entry_points.keys())[0]
    launch_project.get_entry_point(first_entry_point)._validate_parameters(
        launch_project.override_args
    )
    return launch_project



import pytest
import itertools

from typing import List
from ..run import RunState, SweepRun, next_run
from ..config import SweepConfig


def kernel_for_grid_search_tests(
    runs: List[SweepRun], config: SweepConfig, randomize: bool
) -> None:
    """This kernel assumes that sweep config has two categorical parameters
    named v1 and v2."""

    answers = sorted(
        list(
            itertools.product(
                config["parameters"]["v1"]["values"],
                config["parameters"]["v2"]["values"],
            )
        )
    )
    suggested_parameters = [
        (run.config["v1"]["value"], run.config["v2"]["value"]) for run in runs
    ]

    while True:
        suggestion = next_run(config, runs, randomize_order=randomize)
        if suggestion is None:  # done
            break
        assert suggestion.search_info is None
        assert suggestion.state == RunState.pending
        runs.append(suggestion)
        suggested_parameters.append(
            (suggestion.config["v1"]["value"], suggestion.config["v2"]["value"])
        )

    # assert that the grid search iterates over all possible parameters and stops when
    # it exhausts the list of possibilities. do not assert anything about the order
    # in which parameter suggestions are made.
    suggested_parameters = sorted(suggested_parameters)
    assert answers == suggested_parameters


@pytest.mark.parametrize("randomize", [True, False])
def test_grid_from_start_with_and_without_randomize(
    sweep_config_2params_grid_search, randomize
):
    kernel_for_grid_search_tests(
        [], sweep_config_2params_grid_search, randomize=randomize
    )


@pytest.mark.parametrize("randomize", [True, False])
def test_grid_search_starting_from_in_progress(
    sweep_config_2params_grid_search, randomize
):
    runs = [
        SweepRun(config={"v1": {"value": 2}, "v2": {"value": 4}}),
        SweepRun(config={"v1": {"value": 1}, "v2": {"value": 5}}),
    ]
    kernel_for_grid_search_tests(
        runs, sweep_config_2params_grid_search, randomize=randomize
    )



import json
import logging
import os
import shutil
import subprocess
import sys
import tempfile
from typing import Any, Dict, List, Optional, Sequence


from dockerpycreds.utils import find_executable  # type: ignore
import pkg_resources
from six.moves import shlex_quote
import wandb
from wandb.apis.internal import Api
import wandb.docker as docker
from wandb.errors import DockerError, ExecutionError, LaunchError
from wandb.util import get_module
from yaspin import yaspin  # type: ignore

from . import _project_spec
from .utils import _is_wandb_dev_uri, _is_wandb_local_uri
from ..lib.git import GitRepo

_logger = logging.getLogger(__name__)

_GENERATED_DOCKERFILE_NAME = "Dockerfile.wandb-autogenerated"


def validate_docker_installation() -> None:
    """Verify if Docker is installed on host machine."""
    if not find_executable("docker"):
        raise ExecutionError(
            "Could not find Docker executable. "
            "Ensure Docker is installed as per the instructions "
            "at https://docs.docker.com/install/overview/."
        )


def generate_docker_base_image(
    launch_project: _project_spec.LaunchProject, entry_cmd: str
) -> Optional[str]:
    """Uses project and entry point to generate the docker image."""
    path = launch_project.project_dir

    # this check will always pass since the dir attribute will always be populated
    # by _fetch_project_local
    get_module(
        "repo2docker",
        required='wandb launch requires additional dependencies, install with pip install "wandb[launch]"',
    )
    assert isinstance(path, str)
    cmd: Sequence[str] = [
        "jupyter-repo2docker",
        "--no-run",
        "--image-name={}".format(launch_project.base_image),
        "--user-id={}".format(launch_project.docker_user_id),
        path,
        '"{}"'.format(entry_cmd),
    ]

    _logger.info(
        "Generating docker base image from git repo or finding image if it already exists..."
    )
    build_log = os.path.join(launch_project.project_dir, "build.log")
    with yaspin(
        text="Generating docker base image {}, this may take a few minutes...".format(
            launch_project.base_image
        )
    ).bouncingBar.blue as spinner:
        with open(build_log, "w") as f:
            process = subprocess.Popen(cmd, stdout=f, stderr=f)
            res = process.wait()
            if res == 0:
                spinner.text = "Generated docker base image {}".format(
                    launch_project.base_image
                )
                spinner.ok("✅ ")
            else:
                spinner.text = "Detailed error logs: {}".format(build_log)
                spinner.fail("💥 ")
                return None
    return launch_project.base_image


_inspected_images = {}


def docker_image_exists(docker_image: str, should_raise: bool = False) -> bool:
    """Checks if a specific image is already available,
    optionally raising an exception"""
    try:
        data = docker.run(["docker", "image", "inspect", docker_image])
        # always true, since return stderr defaults to false
        assert isinstance(data, str)
        parsed = json.loads(data)[0]
        _inspected_images[docker_image] = parsed
        return True
    except (DockerError, ValueError) as e:
        if should_raise:
            raise e
        return False


def docker_image_inspect(docker_image: str) -> Dict[str, Any]:
    """Get the parsed json result of docker inspect image_name"""
    if _inspected_images.get(docker_image) is None:
        docker_image_exists(docker_image, True)
    return _inspected_images.get(docker_image, {})


def pull_docker_image(docker_image: str) -> None:
    """Pulls the requested docker image"""
    try:
        docker.run(["docker", "pull", docker_image])
    except DockerError as e:
        raise LaunchError("Docker server returned error: {}".format(e))


def build_docker_image_if_needed(
    launch_project: _project_spec.LaunchProject,
    api: Api,
    copy_code: bool,
    workdir: str,
    container_env: List[str],
) -> str:
    """
    Build a docker image containing the project in `work_dir`, using the base image.
    param launch_project: LaunchProject class instance
    :param api: instance of wandb.apis.internal Api
    :param copy_code: boolean indicating if code should be copied into the docker container
    """
    image_uri = _get_docker_image_uri(
        name=launch_project.image_name,
        work_dir=launch_project.project_dir,
        image_id=launch_project.run_id,
    )
    launch_project.docker_image = image_uri
    if docker_image_exists(image_uri) and not launch_project.build_image:
        wandb.termlog("Using existing image: {}".format(image_uri))
        return image_uri
    copy_code_line = ""
    requirements_line = ""
    # TODO: we currently assume the home directory holds the pip cache
    homedir = workdir
    # for custom base_images we attempt to introspect the homedir
    for env in container_env:
        if env.startswith("HOME="):
            homedir = env.split("=", 1)[1]
    if copy_code:
        copy_code_line = "COPY ./src/ {}\n".format(workdir)
        if docker.is_buildx_installed():
            requirements_line = "RUN --mount=type=cache,target={}/.cache,uid={},gid=0 ".format(
                homedir, launch_project.docker_user_id
            )
        else:
            wandb.termwarn(
                "Docker BuildX is not installed, for faster builds upgrade docker: https://github.com/docker/buildx#installing"
            )
            requirements_line = "RUN WANDB_DISABLE_CACHE=true "
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "templates", "_wandb_bootstrap.py"),
            os.path.join(launch_project.project_dir),
        )
        # TODO: make this configurable or change the default behavior...
        requirements_line += _parse_existing_requirements(launch_project)
        requirements_line += "python _wandb_bootstrap.py\n"

    name_line = ""
    if launch_project.name:
        name_line = "ENV WANDB_NAME={wandb_name}\n"
    dockerfile = (
        "FROM {imagename}\n" "{copy_code_line}" "{requirements_line}" "{name_line}"
    ).format(
        imagename=launch_project.base_image,
        copy_code_line=copy_code_line,
        requirements_line=requirements_line,
        name_line=name_line,
    )
    build_ctx_path = _create_docker_build_ctx(launch_project, dockerfile)
    _logger.info("=== Building docker image %s ===", image_uri)

    dockerfile = os.path.join(build_ctx_path, _GENERATED_DOCKERFILE_NAME)
    wandb.termlog("Generating launch image {}".format(image_uri))
    try:
        image = docker.build(
            tags=[image_uri], file=dockerfile, context_path=build_ctx_path
        )
    except DockerError as e:
        raise LaunchError("Error communicating with docker client: {}".format(e))

    try:
        os.remove(build_ctx_path)
    except Exception:
        _logger.info(
            "Temporary docker context file %s was not deleted.", build_ctx_path
        )
    return image


def get_docker_command(
    image: str,
    launch_project: _project_spec.LaunchProject,
    api: Api,
    workdir: str,
    docker_args: Dict[str, Any] = None,
) -> List[str]:
    """Constructs the docker command using the image and docker args.

    Arguments:
    image: a Docker image to be run
    launch_project: an instance of LaunchProject
    api: an instance of wandb.apis.internal Api
    docker_args: a dictionary of additional docker args for the command
    """
    docker_path = "docker"
    cmd: List[Any] = [docker_path, "run", "--rm"]

    if _is_wandb_local_uri(api.settings("base_url")) and sys.platform == "darwin":
        _, _, port = _, _, port = api.settings("base_url").split(":")
        base_url = "http://host.docker.internal:{}".format(port)
    elif _is_wandb_dev_uri(api.settings("base_url")):
        base_url = "http://host.docker.internal:9002"
    else:
        base_url = api.settings("base_url")

    # TODO: only add WANDB_DOCKER when we are pushing the image to a registry
    cmd += [
        "--env",
        f"WANDB_BASE_URL={base_url}",
        "--env",
        f"WANDB_API_KEY={api.api_key}",
        "--env",
        f"WANDB_PROJECT={launch_project.target_project}",
        "--env",
        f"WANDB_ENTITY={launch_project.target_entity}",
        "--env",
        f"WANDB_LAUNCH={True}",
        "--env",
        f"WANDB_LAUNCH_CONFIG_PATH={os.path.join(workdir,_project_spec.DEFAULT_CONFIG_PATH)}",
        "--env",
        f"WANDB_RUN_ID={launch_project.run_id or None}",
        "--env",
        f"WANDB_DOCKER={launch_project.docker_image}",
    ]

    if launch_project.override_config:
        cmd += [
            "-v",
            f"{os.path.join(launch_project.aux_dir, _project_spec.DEFAULT_CONFIG_PATH)}:{os.path.join(workdir,_project_spec.DEFAULT_CONFIG_PATH)}",
        ]
    if docker_args:
        for name, value in docker_args.items():
            # Passed just the name as boolean flag
            if isinstance(value, bool) and value:
                if len(name) == 1:
                    cmd += ["-" + name]
                else:
                    cmd += ["--" + name]
            else:
                # Passed name=value
                if len(name) == 1:
                    cmd += ["-" + name, value]
                else:
                    cmd += ["--" + name, value]

    cmd += [image]
    return [shlex_quote(c) for c in cmd]


def _parse_existing_requirements(launch_project: _project_spec.LaunchProject) -> str:
    requirements_line = ""
    base_requirements = os.path.join(launch_project.project_dir, "requirements.txt")
    if os.path.exists(base_requirements):
        include_only = set()
        with open(base_requirements) as f:
            iter = pkg_resources.parse_requirements(f)
            while True:
                try:
                    pkg = next(iter)
                    if hasattr(pkg, "name"):
                        name = pkg.name.lower()  # type: ignore
                    else:
                        name = str(pkg)
                    include_only.add(shlex_quote(name))
                    requirements_line += "WANDB_ONLY_INCLUDE={} ".format(
                        ",".join(include_only)
                    )
                except StopIteration:
                    break
                # Different versions of pkg_resources throw different errors
                # just catch them all and ignore packages we can't parse
                except Exception as e:
                    _logger.warn(f"Unable to parse requirements.txt: {e}")
                    continue
    return requirements_line


def _get_docker_image_uri(name: Optional[str], work_dir: str, image_id: str) -> str:
    """
    Returns an appropriate Docker image URI for a project based on the git hash of the specified
    working directory.
    :param name: The URI of the Docker repository with which to tag the image. The
                           repository URI is used as the prefix of the image URI.
    :param work_dir: Path to the working directory in which to search for a git commit hash
    """
    name = name.replace(" ", "-") if name else "wandb-launch"
    # Optionally include first 7 digits of git SHA in tag name, if available.

    git_commit = GitRepo(work_dir).last_commit
    version_string = ":" + str(git_commit[:7]) + image_id if git_commit else image_id
    return name + version_string


def _create_docker_build_ctx(
    launch_project: _project_spec.LaunchProject, dockerfile_contents: str,
) -> str:
    """Creates build context temp dir containing Dockerfile and project code, returning path to temp dir."""
    directory = tempfile.mkdtemp()
    dst_path = os.path.join(directory, "src")
    shutil.copytree(src=launch_project.project_dir, dst=dst_path)
    if launch_project.python_version:
        runtime_path = os.path.join(dst_path, "runtime.txt")
        with open(runtime_path, "w") as fp:
            fp.write(f"python-{launch_project.python_version}")
    # TODO: we likely don't need to pass the whole git repo into the container
    # with open(os.path.join(directory, ".dockerignore"), "w") as f:
    #    f.write("**/.git")
    with open(os.path.join(directory, _GENERATED_DOCKERFILE_NAME), "w") as handle:
        handle.write(dockerfile_contents)
    return directory



import logging
import sys
from typing import Any, Dict, List, Optional

from wandb.apis.internal import Api
from wandb.errors import ExecutionError, LaunchError

from ._project_spec import create_project_from_spec, fetch_and_validate_project
from .agent import LaunchAgent
from .runner import loader
from .runner.abstract import AbstractRun
from .utils import (
    _is_wandb_local_uri,
    construct_launch_spec,
    PROJECT_DOCKER_ARGS,
    PROJECT_SYNCHRONOUS,
)

_logger = logging.getLogger(__name__)


def run_agent(entity: str, project: str, queues: Optional[List[str]] = None) -> None:
    agent = LaunchAgent(entity, project, queues)
    agent.loop()


def _run(
    uri: str,
    name: Optional[str],
    project: Optional[str],
    entity: Optional[str],
    docker_image: Optional[str],
    entry_point: Optional[str],
    version: Optional[str],
    parameters: Optional[Dict[str, Any]],
    docker_args: Optional[Dict[str, Any]],
    resource: str,
    launch_config: Optional[Dict[str, Any]],
    synchronous: Optional[bool],
    api: Api,
) -> AbstractRun:
    """Helper that delegates to the project-running method corresponding to the passed-in backend."""
    launch_spec = construct_launch_spec(
        uri,
        api,
        name,
        project,
        entity,
        docker_image,
        entry_point,
        version,
        parameters,
        launch_config,
    )
    launch_project = create_project_from_spec(launch_spec, api)
    launch_project = fetch_and_validate_project(launch_project, api)

    # construct runner config.
    runner_config: Dict[str, Any] = {}
    runner_config[PROJECT_SYNCHRONOUS] = synchronous
    runner_config[PROJECT_DOCKER_ARGS] = docker_args

    backend = loader.load_backend(resource, api, runner_config)
    if backend:
        submitted_run = backend.run(launch_project)
        # this check will always pass, run is only optional in the agent case where
        # a run queue id is present on the backend config
        assert submitted_run
        return submitted_run
    else:
        raise ExecutionError(
            "Unavailable backend {}, available backends: {}".format(
                resource, ", ".join(loader.WANDB_RUNNERS.keys())
            )
        )


def run(
    uri: str,
    api: Api,
    entry_point: Optional[str] = None,
    version: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
    docker_args: Optional[Dict[str, Any]] = None,
    name: Optional[str] = None,
    resource: str = "local",
    project: Optional[str] = None,
    entity: Optional[str] = None,
    docker_image: Optional[str] = None,
    config: Optional[Dict[str, Any]] = None,
    synchronous: Optional[bool] = True,
) -> AbstractRun:
    """Run a W&B launch experiment. The project can be wandb uri or a Git URI.

    Arguments:
    uri: URI of experiment to run. A wandb run uri or a Git repository URI.
    api: An instance of a wandb Api from wandb.apis.internal.
    entry_point: Entry point to run within the project. Defaults to using the entry point used
        in the original run for wandb URIs, or main.py for git repository URIs.
    version: For Git-based projects, either a commit hash or a branch name.
    parameters: Parameters (dictionary) for the entry point command. Defaults to using the
        the parameters used to run the original run.
    docker_args: Arguments (dictionary) for the docker command.
    name: Name run under which to launch the run.
    resource: Execution backend for the run: W&B provides built-in support for "local" backend
    project: Target project to send launched run to
    entity: Target entity to send launched run to
    config: A dictionary which will be passed as config to the backend. The exact content
        which should be provided is different for each execution backend
    synchronous: Whether to block while waiting for a run to complete. Defaults to True.
        Note that if ``synchronous`` is False and ``backend`` is "local", this
        method will return, but the current process will block when exiting until
        the local run completes. If the current process is interrupted, any
        asynchronous runs launched via this method will be terminated. If
        ``synchronous`` is True and the run fails, the current process will
        error out as well.


    Example:
        import wandb
        project_uri = "https://github.com/wandb/examples"
        params = {"alpha": 0.5, "l1_ratio": 0.01}
        # Run W&B project and create a reproducible docker environment
        # on a local host
        api = wandb.apis.internal.Api()
        wandb.launch(project_uri, api, parameters=params)


    Returns:
        an instance of`wandb.launch.SubmittedRun` exposing information (e.g. run ID)
        about the launched run.

    Raises:
        `wandb.exceptions.ExecutionError` If a run launched in blocking mode
        is unsuccessful.
    """
    if docker_args is None:
        docker_args = {}

    if _is_wandb_local_uri(api.settings("base_url")):
        if sys.platform == "win32":
            docker_args["net"] = "host"
        else:
            docker_args["network"] = "host"
        if sys.platform == "linux" or sys.platform == "linux2":
            docker_args["add-host"] = "host.docker.internal:host-gateway"

    if config is None:
        config = {}

    submitted_run_obj = _run(
        uri=uri,
        name=name,
        project=project,
        entity=entity,
        docker_image=docker_image,
        entry_point=entry_point,
        version=version,
        parameters=parameters,
        docker_args=docker_args,
        resource=resource,
        launch_config=config,
        synchronous=synchronous,
        api=api,
    )

    if synchronous:
        _wait_for(submitted_run_obj)
    else:
        raise LaunchError("Non synchronous mode not supported")
    return submitted_run_obj


def _wait_for(submitted_run_obj: AbstractRun) -> None:
    """Wait on the passed-in submitted run, reporting its status to the tracking server."""
    # Note: there's a small chance we fail to report the run's status to the tracking server if
    # we're interrupted before we reach the try block below
    try:
        if submitted_run_obj.wait():
            _logger.info("=== Submitted run succeeded ===")
        else:
            raise ExecutionError("Submitted run failed")
    except KeyboardInterrupt:
        _logger.error("=== Submitted run interrupted, cancelling run ===")
        submitted_run_obj.cancel()
        raise



import pytest

from ..config import SweepConfig
import numpy as np
from ..run import next_run
from .._types import ArrayLike
import os
from scipy import stats

from pathlib import Path

test_results_dir = Path(__file__).parent.parent / "test_results"
test_results_dir.mkdir(parents=True, exist_ok=True)


def check_that_samples_are_from_the_same_distribution(
    pred_samples,
    true_samples,
    bins,
):

    n_pred, _ = np.histogram(pred_samples, bins=bins)
    n_true, _ = np.histogram(true_samples, bins=bins)

    # assert the counts are equal in each bin to 1 sigma within the poisson error
    err_pred = np.sqrt(n_pred)
    err_true = np.sqrt(n_true)

    # less than 5 sigma different

    # if denom is zero, then both bins have zero counts, to set to 1 to
    # avoid division by zero error (no effect on answer)
    denom = np.sqrt(err_pred ** 2 + err_true ** 2)
    denom[np.isclose(denom, 0)] = 1
    sigma_diff = np.abs(n_pred - n_true) / denom
    sigma_diff[~np.isfinite(sigma_diff)] = 0

    np.testing.assert_array_less(sigma_diff, 5)


def plot_two_distributions(
    samples_true: ArrayLike,
    samples_pred: ArrayLike,
    bins: ArrayLike,
    xscale="linear",
):
    import matplotlib.pyplot as plt
    import inspect

    fig, ax = plt.subplots()
    ax.hist(
        samples_true,
        bins=bins,
        histtype="stepfilled",
        label="true",
        alpha=0.2,
    )
    ax.hist(
        samples_pred,
        bins=bins,
        histtype="stepfilled",
        label="pred",
        alpha=0.2,
    )
    ax.legend()
    ax.set_xscale(xscale)
    ax.tick_params(which="both", axis="both", direction="in")
    current_test = os.environ.get("PYTEST_CURRENT_TEST")
    if current_test is None:
        current_test = inspect.stack()[1].function
    else:
        current_test = current_test.split(":")[-1].split(" ")[0]
    fname = f"{current_test}.pdf"
    fig.savefig(test_results_dir / fname)


def test_rand_uniform(plot):

    v1_min = 3.0
    v1_max = 5.0
    n_samples = 1000

    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"min": v1_min, "max": v1_max},
            },
        }
    )

    runs = []
    for i in range(n_samples):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.random.uniform(v1_min, v1_max, size=n_samples)
    bins = np.linspace(v1_min, v1_max, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins)

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)


def test_rand_normal(plot):
    # Calculates that the

    n_samples = 1000

    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "normal"},
            },
        }
    )

    runs = []
    for i in range(n_samples):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.random.normal(0, 1, size=n_samples)
    bins = np.linspace(-2, 2, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins)

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)


def test_rand_lognormal(plot):
    # Calculates that the

    n_samples = 1000

    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "log_normal", "mu": 2, "sigma": 3},
            },
        }
    )

    runs = []
    for i in range(n_samples):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.random.lognormal(2, 3, size=n_samples)

    bins = np.logspace(-1, 5, 30)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins, xscale="log")

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)


def test_rand_loguniform(plot):
    # Calculates that the

    v2_min = 5.0
    v2_max = 100
    n_samples = 1000

    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v2": {
                    "min": np.log(v2_min),
                    "max": np.log(v2_max),
                    "distribution": "log_uniform",
                },
            },
        }
    )

    runs = []
    for i in range(n_samples):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v2"]["value"] for run in runs])
    true_samples = np.random.uniform(np.log(v2_min), np.log(v2_max), size=n_samples)
    true_samples = np.exp(true_samples)

    # the lhs needs to be >= 0 because
    bins = np.logspace(np.log10(v2_min), np.log10(v2_max), 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins, xscale="log")

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)

    assert pred_samples.min() >= v2_min
    assert pred_samples.max() <= v2_max


@pytest.mark.parametrize("q", [0.1, 1, 10])
def test_rand_q_lognormal(q, plot):

    n_samples_true = 1000
    n_samples_pred = 1000
    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "q_log_normal", "mu": 2, "sigma": 2, "q": q},
            },
        }
    )

    runs = []
    for i in range(n_samples_pred):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.round(np.random.lognormal(2, 2, size=n_samples_true) / q) * q

    # need the binsize to be >> q
    bins = np.logspace(np.log10(np.exp(-2)), np.log10(np.exp(6)), 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins, xscale="log")

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)

    remainder = np.remainder(pred_samples, q)

    # when pred_samples == 0, pred_samples % q = q, so need to test for both remainder = q and
    # remainder = 0 under modular division
    assert np.all(np.isclose(remainder, 0) | np.isclose(remainder, q))


@pytest.mark.parametrize("q", [0.1, 1, 10])
def test_rand_q_normal(q, plot):

    n_samples_true = 1000
    n_samples_pred = 1000
    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "q_normal", "mu": 4, "sigma": 2, "q": q},
            },
        }
    )

    runs = []
    for i in range(n_samples_pred):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.round(np.random.normal(4, 2, size=n_samples_true) / q) * q

    # need the binsize to be >> q
    bins = np.linspace(0, 8, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins)

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)
    remainder = np.remainder(pred_samples, q)

    # when pred_samples == 0, pred_samples % q = q, so need to test for both remainder = q and
    # remainder = 0 under modular division
    assert np.all(np.isclose(remainder, 0) | np.isclose(remainder, q))


@pytest.mark.parametrize("q", [0.1, 1, 10])
def test_rand_q_uniform(q, plot):

    n_samples_true = 1000
    n_samples_pred = 1000
    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "q_uniform", "min": 0, "max": 100, "q": q},
            },
        }
    )

    runs = []
    for i in range(n_samples_pred):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.round(np.random.uniform(0, 100, size=n_samples_true) / q) * q

    # need the binsize to be >> q
    bins = np.linspace(0, 100, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins)

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)
    remainder = np.remainder(pred_samples, q)

    # when pred_samples == 0, pred_samples % q = q, so need to test for both remainder = q and
    # remainder = 0 under modular division
    assert np.all(np.isclose(remainder, 0) | np.isclose(remainder, q))


@pytest.mark.parametrize("q", [0.1, 1, 10])
def test_rand_q_loguniform(q, plot):

    n_samples_pred = 1000
    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {
                    "distribution": "q_log_uniform",
                    "min": np.log(0.1),
                    "max": np.log(100),
                    "q": q,
                },
            },
        }
    )

    runs = []
    for i in range(n_samples_pred):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.round(stats.loguniform(0.1, 100).rvs(1000) / q) * q

    # need the binsize to be >> q
    bins = np.logspace(-1, 2, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins, xscale="log")

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)
    remainder = np.remainder(pred_samples, q)

    # when pred_samples == 0, pred_samples % q = q, so need to test for both remainder = q and
    # remainder = 0 under modular division
    assert np.all(np.isclose(remainder, 0) | np.isclose(remainder, q))


@pytest.mark.parametrize("q", [0.1])
def test_rand_q_beta(q, plot):

    n_samples_pred = 1000
    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "q_beta", "a": 2, "b": 5, "q": q},
            },
        }
    )

    runs = []
    for i in range(n_samples_pred):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.round(np.random.beta(2, 5, 1000) / q) * q

    # need the binsize to be >> q
    bins = np.linspace(0, 1, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins)

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)
    remainder = np.remainder(pred_samples, q)

    # when pred_samples == 0, pred_samples % q = q, so need to test for both remainder = q and
    # remainder = 0 under modular division
    assert np.all(np.isclose(remainder, 0) | np.isclose(remainder, q))


def test_rand_beta(plot):

    n_samples_pred = 1000
    sweep_config_2params = SweepConfig(
        {
            "method": "random",
            "parameters": {
                "v1": {"distribution": "beta", "a": 2, "b": 5},
            },
        }
    )

    runs = []
    for i in range(n_samples_pred):
        suggestion = next_run(sweep_config_2params, runs)
        runs.append(suggestion)

    pred_samples = np.asarray([run.config["v1"]["value"] for run in runs])
    true_samples = np.random.beta(2, 5, 1000)

    # need the binsize to be >> q
    bins = np.linspace(0, 1, 10)

    if plot:
        plot_two_distributions(true_samples, pred_samples, bins)

    check_that_samples_are_from_the_same_distribution(pred_samples, true_samples, bins)



import pytest
from ..config import SweepConfig


def pytest_addoption(parser):
    parser.addoption(
        "--plot",
        action="store_true",
        help="Plot true and predicted distributions "
        "for tests involving random sampling.",
    )


@pytest.fixture
def plot(request):
    return request.config.getoption("--plot")


@pytest.fixture()
def sweep_config_bayes_search_2params_with_metric():
    return SweepConfig(
        {
            "metric": {"name": "loss", "goal": "minimize"},
            "method": "bayes",
            "parameters": {
                "v1": {"min": 1, "max": 10},
                "v2": {"min": 1.0, "max": 10.0},
            },
        }
    )


@pytest.fixture()
def sweep_config_2params_grid_search():
    return SweepConfig(
        {
            "method": "grid",
            "parameters": {"v1": {"values": [1, 2, 3]}, "v2": {"values": [4, 5]}},
        }
    )



import json
import logging
import os
import re
import signal
import subprocess
import sys
from typing import Any, Dict, List, Optional

import wandb
from wandb.errors import CommError, LaunchError

from .abstract import AbstractRun, AbstractRunner, Status
from .._project_spec import DEFAULT_CONFIG_PATH, get_entry_point_command, LaunchProject
from ..docker import (
    build_docker_image_if_needed,
    docker_image_exists,
    docker_image_inspect,
    generate_docker_base_image,
    get_docker_command,
    pull_docker_image,
    validate_docker_installation,
)
from ..utils import (
    PROJECT_DOCKER_ARGS,
    PROJECT_SYNCHRONOUS,
)


_logger = logging.getLogger(__name__)


class LocalSubmittedRun(AbstractRun):
    """Instance of ``AbstractRun`` corresponding to a subprocess launched to run an entry point command locally."""

    def __init__(self, command_proc: "subprocess.Popen[bytes]") -> None:
        super().__init__()
        self.command_proc = command_proc

    @property
    def id(self) -> int:
        return self.command_proc.pid

    def wait(self) -> bool:
        return self.command_proc.wait() == 0

    def cancel(self) -> None:
        # Interrupt child process if it hasn't already exited
        if self.command_proc.poll() is None:
            # Kill the the process tree rooted at the child if it's the leader of its own process
            # group, otherwise just kill the child
            try:
                if self.command_proc.pid == os.getpgid(self.command_proc.pid):
                    os.killpg(self.command_proc.pid, signal.SIGTERM)
                else:
                    self.command_proc.terminate()
            except OSError:
                # The child process may have exited before we attempted to terminate it, so we
                # ignore OSErrors raised during child process termination
                _logger.info(
                    "Failed to terminate child process (PID %s). The process may have already exited.",
                    self.command_proc.pid,
                )
            self.command_proc.wait()

    def get_status(self) -> Status:
        exit_code = self.command_proc.poll()
        if exit_code is None:
            return Status("running")
        if exit_code == 0:
            return Status("finished")
        return Status("failed")


class LocalRunner(AbstractRunner):
    """Runner class, uses a project to create a LocallySubmittedRun"""

    def run(self, launch_project: LaunchProject) -> Optional[AbstractRun]:
        validate_docker_installation()
        synchronous: bool = self.backend_config[PROJECT_SYNCHRONOUS]
        docker_args: Dict[str, Any] = self.backend_config[PROJECT_DOCKER_ARGS]

        entry_point = launch_project.get_single_entry_point()

        entry_cmd = entry_point.command
        copy_code = True
        if launch_project.docker_image:
            pull_docker_image(launch_project.docker_image)
            copy_code = False
        else:
            # TODO: potentially pull the base_image
            if not docker_image_exists(launch_project.base_image):
                if generate_docker_base_image(launch_project, entry_cmd) is None:
                    raise LaunchError("Unable to build base image")
            else:
                wandb.termlog(
                    "Using existing base image: {}".format(launch_project.base_image)
                )

        command_args = []
        command_separator = " "

        container_inspect = docker_image_inspect(launch_project.base_image)
        container_workdir = container_inspect["ContainerConfig"].get("WorkingDir", "/")
        container_env: List[str] = container_inspect["ContainerConfig"]["Env"]

        if launch_project.docker_image is None or launch_project.build_image:
            image = build_docker_image_if_needed(
                launch_project=launch_project,
                api=self._api,
                copy_code=copy_code,
                workdir=container_workdir,
                container_env=container_env,
            )
        else:
            image = launch_project.docker_image
        command_args += get_docker_command(
            image=image,
            launch_project=launch_project,
            api=self._api,
            workdir=container_workdir,
            docker_args=docker_args,
        )
        if self.backend_config.get("runQueueItemId"):
            try:
                self._api.ack_run_queue_item(
                    self.backend_config["runQueueItemId"], launch_project.run_id
                )
            except CommError:
                wandb.termerror(
                    "Error acking run queue item. Item lease may have ended or another process may have acked it."
                )
                return None

        # In synchronous mode, run the entry point command in a blocking fashion, sending status
        # updates to the tracking server when finished. Note that the run state may not be
        # persisted to the tracking server if interrupted
        if synchronous:
            command_args += get_entry_point_command(
                entry_point, launch_project.override_args
            )
            if launch_project.override_config:
                with open(
                    os.path.join(launch_project.aux_dir, DEFAULT_CONFIG_PATH), "w"
                ) as fp:
                    json.dump(launch_project.override_config, fp)
            command_str = command_separator.join(command_args)

            wandb.termlog(
                "Launching run in docker with command: {}".format(
                    re.sub(r"WANDB_API_KEY=\w+", "WANDB_API_KEY", command_str)
                )
            )
            run = _run_entry_point(command_str, launch_project.project_dir)
            run.wait()
            return run
        # Otherwise, invoke `wandb launch` in a subprocess
        raise LaunchError("asynchronous mode not yet available")


def _run_launch_cmd(cmd: List[str]) -> "subprocess.Popen[str]":
    """Invoke ``wandb launch`` in a subprocess, which in turn runs the entry point in a child process.

    Arguments:
    cmd: List of strings indicating the command to run

    Returns:
        A handle to the subprocess. Popen launched to invoke ``wandb launch``.
    """
    final_env = os.environ.copy()
    # Launch `wandb launch` command as the leader of its own process group so that we can do a
    # best-effort cleanup of all its descendant processes if needed
    if sys.platform == "win32":
        return subprocess.Popen(
            cmd,
            env=final_env,
            universal_newlines=True,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
        )
    else:
        return subprocess.Popen(
            cmd, env=final_env, universal_newlines=True, preexec_fn=os.setsid
        )


def _run_entry_point(command: str, work_dir: str) -> AbstractRun:
    """Run an entry point command in a subprocess.

    Arguments:
    command: Entry point command to run
    work_dir: Working directory in which to run the command
    run: SubmittedRun object associated with the entry point execution.

    Returns:
        An instance of `LocalSubmittedRun`
    """
    env = os.environ.copy()
    if os.name == "nt":
        # we are running on windows
        process = subprocess.Popen(
            ["cmd", "/c", command], close_fds=True, cwd=work_dir, env=env
        )
    else:
        process = subprocess.Popen(
            ["bash", "-c", command], close_fds=True, cwd=work_dir, env=env,
        )

    return LocalSubmittedRun(process)



import logging
from typing import Any, Dict

from wandb.apis.internal import Api
from wandb.errors import LaunchError
from wandb.sdk.launch.runner.abstract import AbstractRunner

from .local import LocalRunner

__logger__ = logging.getLogger(__name__)


# Statically register backend defined in wandb
WANDB_RUNNERS = {"local": LocalRunner}


def load_backend(
    backend_name: str, api: Api, backend_config: Dict[str, Any]
) -> AbstractRunner:
    # Static backends
    if backend_name in WANDB_RUNNERS:
        return WANDB_RUNNERS[backend_name](api, backend_config)

    raise LaunchError(
        "Resource name not among available resources. Available resources: {} ".format(
            ",".join(list(WANDB_RUNNERS.keys()))
        )
    )



"""
Implementation of launch agent.
"""

import os
import sys
import time
from typing import Any, Dict, Iterable, List, Optional

import wandb
from wandb.apis.internal import Api
from wandb.errors import LaunchError
import wandb.util as util

from .._project_spec import create_project_from_spec, fetch_and_validate_project
from ..runner.abstract import AbstractRun, State
from ..runner.loader import load_backend
from ..utils import (
    _is_wandb_local_uri,
    PROJECT_DOCKER_ARGS,
    PROJECT_SYNCHRONOUS,
)

AGENT_POLLING_INTERVAL = 10


def _convert_access(access: str) -> str:
    """Converts access string to a value accepted by wandb."""
    access = access.upper()
    assert (
        access == "PROJECT" or access == "USER"
    ), "Queue access must be either project or user"
    return access


class LaunchAgent(object):
    """Launch agent class which polls run given run queues and launches runs for wandb launch."""

    STATE_MAP: Dict[str, State] = {}

    def __init__(self, entity: str, project: str, queues: Iterable[str] = None):
        self._entity = entity
        self._project = project
        self._api = Api()
        self._settings = wandb.Settings()
        self._base_url = self._api.settings().get("base_url")
        self._jobs: Dict[int, AbstractRun] = {}
        self._ticks = 0
        self._running = 0
        self._cwd = os.getcwd()
        self._namespace = wandb.util.generate_id()
        self._access = _convert_access("project")
        self._queues: List[str] = []
        self.setup_run_queues(queues)

    def setup_run_queues(self, queues: Optional[Iterable[str]]) -> None:
        """Checks the project to ensure run queues exist then adds them to a list to be watched by the agent."""
        # TODO: add run queue filtering to server
        project_run_queues = self._api.get_project_run_queues(
            self._entity, self._project
        )
        existing_run_queue_names = set(
            [run_queue["name"] for run_queue in project_run_queues]
        )
        if queues is None:
            for queue in project_run_queues:
                if queue["name"] == "default":
                    self._queues = ["default"]
                    return
            raise LaunchError(
                "Error launching launch-agent: Requested default queue for {}/{} but default queue does not exist.".format(
                    self._entity, self._project
                )
            )
        else:
            for queue in queues:
                if queue not in existing_run_queue_names:
                    raise LaunchError(
                        "Error launching launch-agent: {} does not exist for {}/{}".format(
                            queue, self._entity, self._project
                        )
                    )
                self._queues.append(queue)

    @property
    def job_ids(self) -> List[int]:
        """Returns a list of keys running job ids for the agent."""
        return list(self._jobs.keys())

    def pop_from_queue(self, queue: str) -> Any:
        """Pops an item off the runqueue to run as a job."""
        try:
            ups = self._api.pop_from_run_queue(
                queue, entity=self._entity, project=self._project
            )
        except Exception as e:
            print("Exception:", e)
            return None
        return ups

    def print_status(self) -> None:
        """Prints the current status of the agent."""
        print(
            "polling on project {}, queues {} for jobs".format(
                self._project, " ".join(self._queues)
            )
        )

    def finish_job_id(self, job_id: int) -> None:
        """Removes the job from our list for now."""
        # TODO:  keep logs or something for the finished jobs
        del self._jobs[job_id]
        self._running -= 1

    def _update_finished(self, job_id: int) -> None:
        """Check our status enum."""
        if self._jobs[job_id].get_status().state in ["failed", "finished"]:
            self.finish_job_id(job_id)

    def run_job(self, job: Dict[str, Any]) -> None:
        """Sets up project and runs the job."""
        # TODO: logger
        print("agent: got job", job)
        # parse job
        launch_spec = job["runSpec"]
        if launch_spec.get("overrides") and isinstance(
            launch_spec["overrides"].get("args"), list
        ):
            launch_spec["overrides"]["args"] = util._user_args_to_dict(
                launch_spec["overrides"].get("args", [])
            )
        project = create_project_from_spec(launch_spec, self._api)
        project = fetch_and_validate_project(project, self._api)

        resource = launch_spec.get("resource") or "local"
        backend_config: Dict[str, Any] = {
            PROJECT_DOCKER_ARGS: {},
            PROJECT_SYNCHRONOUS: True,
        }
        if _is_wandb_local_uri(self._base_url):
            if sys.platform == "win32":
                backend_config[PROJECT_DOCKER_ARGS]["net"] = "host"
            else:
                backend_config[PROJECT_DOCKER_ARGS]["network"] = "host"
            if sys.platform == "linux" or sys.platform == "linux2":
                backend_config[PROJECT_DOCKER_ARGS][
                    "add-host"
                ] = "host.docker.internal:host-gateway"

        backend_config["runQueueItemId"] = job["runQueueItemId"]
        backend = load_backend(resource, self._api, backend_config)
        backend.verify()

        run = backend.run(project)
        if run:
            self._jobs[run.id] = run
            self._running += 1

    def loop(self) -> None:
        """Main loop function for agent."""
        wandb.termlog(
            "launch agent polling project {}/{} on queues: {}".format(
                self._entity, self._project, ",".join(self._queues)
            )
        )
        try:
            while True:
                self._ticks += 1
                for queue in self._queues:
                    job = self.pop_from_queue(queue)
                    if job:
                        break
                if not job:
                    time.sleep(AGENT_POLLING_INTERVAL)
                    # for job_id in self.job_ids:
                    # self._update_finished(job_id)
                    if self._ticks % 2 == 0:
                        self.print_status()
                    continue
                self.run_job(job)
        except KeyboardInterrupt:
            wandb.termlog("Shutting down, active jobs:")
            self.print_status()



from typing import List, Optional, Union, Any, Dict
from enum import Enum
import numpy as np
import datetime

from pydantic import BaseModel, Field
from .config import SweepConfig
from ._types import floating


class RunState(str, Enum):
    pending = "pending"
    running = "running"
    finished = "finished"
    killed = "killed"
    crashed = "crashed"
    failed = "failed"
    preempted = "preempted"
    preempting = "preempting"


class SweepRun(BaseModel):
    """A wandb Run that is part of a Sweep.

    >>> run = SweepRun(
    ...   name="my_run",
    ...   state=RunState.running,
    ...   config={"a": {"value": 1}},
    ... )

    Args:
        name: Name of the run.
        state: State of the run.
        config: `dict` representation of the run's wandb.config.
        summaryMetrics: `dict` of summary statistics for the run.
        history: List of dicts containing the arguments to calls of wandb.log made during the run.
        search_info: Dict containing information produced by the search algorithm.
        early_terminate_info: Dict containing information produced by the early terminate algorithm.
        stopped: Whether the run was stopped in the sweep
        shouldStop: Whether the run should stop in the sweep
        heartbeat_at: The last time the backend received a heart beat from the run
        exitcode: The exitcode of the process that trained the run
        running: Whether the run is currently running
    """

    name: Optional[str] = None
    summary_metrics: Optional[dict] = Field(
        default_factory=lambda: {}, alias="summaryMetrics"
    )
    history: List[dict] = Field(default_factory=lambda: [], alias="sampledHistory")
    config: dict = Field(default_factory=lambda: {})
    state: RunState = RunState.pending
    search_info: Optional[Dict] = None
    early_terminate_info: Optional[Dict] = None
    stopped: bool = False
    should_stop: bool = Field(default=False, alias="shouldStop")
    heartbeat_at: Optional[datetime.datetime] = Field(default=None, alias="heartbeatAt")
    exitcode: Optional[int] = None
    running: Optional[bool] = None

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True

    def metric_history(self, metric_name: str) -> List[floating]:
        return [d[metric_name] for d in self.history if metric_name in d]

    def summary_metric(self, metric_name: str) -> floating:
        if self.summary_metrics is None:
            raise ValueError("this run has no summary metrics")
        if metric_name not in self.summary_metrics:
            raise KeyError(f"{metric_name} is not a summary metric of this run.")
        return self.summary_metrics[metric_name]

    def metric_extremum(self, metric_name: str, kind: str) -> floating:
        """Calculate the maximum or minimum value of a specified metric.

        >>> run = SweepRun(history=[{'a': 1}, {'b': 3}, {'a': 2, 'b': 4}], summary_metrics={'a': 50})
        >>> assert run.metric_extremum('a', 'maximum') == 50

        Args:
            metric_name: The name of the target metric.
            kind: What kind of extremum to get (either "maximum" or "minimum").

        Returns:
            The maximum or minimum metric.
        """

        cmp_func = np.max if kind == "maximum" else np.min
        try:
            summary_metric = [self.summary_metric(metric_name)]
        except KeyError:
            summary_metric = []
        all_metrics = self.metric_history(metric_name) + summary_metric

        if len(all_metrics) == 0:
            raise ValueError(f"Cannot extract metric {metric_name} from run")

        def filter_func(x: Any) -> bool:
            try:
                return np.isscalar(x) and np.isfinite(x)
            except TypeError:
                return False

        all_metrics = list(filter(filter_func, all_metrics))

        if len(all_metrics) == 0:
            raise ValueError("Run does not have any finite metric values")

        return cmp_func(all_metrics)


def next_run(
    sweep_config: Union[dict, SweepConfig],
    runs: List[SweepRun],
    validate: bool = False,
    **kwargs,
) -> Optional[SweepRun]:
    """Calculate the next run in a sweep.

    >>> suggested_run = next_run({
    ...    'method': 'grid',
    ...    'parameters': {'a': {'values': [1, 2, 3]}}
    ... }, [])
    >>> assert suggested_run.config['a']['value'] == 1

    Args:
        sweep_config: The config for the sweep.
        runs: List of runs in the sweep.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.

    Returns:
        The suggested run.
    """

    from .grid_search import grid_search_next_run
    from .random_search import random_search_next_run
    from .bayes_search import bayes_search_next_run

    # validate the sweep config
    if validate:
        sweep_config = SweepConfig(sweep_config)

    # this access is safe due to the jsonschema
    method = sweep_config["method"]

    if method == "grid":
        return grid_search_next_run(runs, sweep_config, validate=validate, **kwargs)
    elif method == "random":
        return random_search_next_run(sweep_config, validate=validate)
    elif method == "bayes":
        return bayes_search_next_run(runs, sweep_config, validate=validate, **kwargs)
    else:
        raise ValueError(
            f'Invalid search type {method}, must be one of ["grid", "random", "bayes"]'
        )


def stop_runs(
    sweep_config: Union[dict, SweepConfig],
    runs: List[SweepRun],
    validate: bool = False,
) -> List[SweepRun]:
    """Calculate the runs in a sweep to stop by early termination.

    >>> to_stop = stop_runs({
    ...    "method": "grid",
    ...    "metric": {"name": "loss", "goal": "minimize"},
    ...    "early_terminate": {
    ...        "type": "hyperband",
    ...        "max_iter": 5,
    ...        "eta": 2,
    ...        "s": 2,
    ...    },
    ...    "parameters": {"a": {"values": [1, 2, 3]}},
    ... }, [
    ...    SweepRun(
    ...        name="a",
    ...        state=RunState.finished,  # This is already stopped
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 9},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="b",
    ...        state=RunState.running,  # This should be stopped
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 10},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="c",
    ...        state=RunState.running,  # This passes band 1 but not band 2
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 8},
    ...            {"loss": 8},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="d",
    ...        state=RunState.running,
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 7},
    ...            {"loss": 7},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="e",
    ...        state=RunState.finished,
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 6},
    ...            {"loss": 6},
    ...        ],
    ...    ),
    ...])


    Args:
        sweep_config: The config for the sweep.
        runs: List of runs in the sweep.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.


    Returns:
        A list of the runs to stop.
    """

    from .hyperband_stopping import hyperband_stop_runs

    # validate the sweep config
    if validate:
        sweep_config = SweepConfig(sweep_config)

    if "metric" not in sweep_config:
        raise ValueError('early terminate requires "metric" section')

    if "early_terminate" not in sweep_config:
        raise ValueError('early terminate requires "early_terminate" section.')
    et_type = sweep_config["early_terminate"]["type"]

    if et_type == "hyperband":
        return hyperband_stop_runs(runs, sweep_config, validate=validate)
    else:
        raise ValueError(
            f'Invalid early stopping type {et_type}, must be one of ["hyperband"]'
        )



import numpy as np

from typing import List, Tuple, Optional, Union

from .config.cfg import SweepConfig
from .run import SweepRun, RunState
from .params import HyperParameter, HyperParameterSet
from sklearn import gaussian_process as sklearn_gaussian
from scipy import stats as scipy_stats

from ._types import floating, integer, ArrayLike

NUGGET = 1e-10


def fit_normalized_gaussian_process(
    X: ArrayLike, y: ArrayLike, nu: floating = 1.5
) -> Tuple[sklearn_gaussian.GaussianProcessRegressor, floating, floating]:
    gp = sklearn_gaussian.GaussianProcessRegressor(
        kernel=sklearn_gaussian.kernels.Matern(nu=nu),
        n_restarts_optimizer=2,
        alpha=0.0000001,
        random_state=2,
    )

    y_stddev: ArrayLike
    if len(y) == 1:
        y = np.array(y)
        y_mean = y[0]
        y_stddev = 1.0
    else:
        y_mean = np.mean(y)
        y_stddev = np.std(y) + 0.0001
    y_norm = (y - y_mean) / y_stddev
    gp.fit(X, y_norm)
    return gp, y_mean, y_stddev


def sigmoid(x: ArrayLike) -> ArrayLike:
    return np.exp(-np.logaddexp(0, -x))


def random_sample(X_bounds: ArrayLike, num_test_samples: integer) -> ArrayLike:
    num_hyperparameters = len(X_bounds)
    test_X = np.empty((int(num_test_samples), num_hyperparameters))
    for ii in range(num_test_samples):
        for jj in range(num_hyperparameters):
            if type(X_bounds[jj][0]) == int:
                assert type(X_bounds[jj][1]) == int
                test_X[ii, jj] = np.random.randint(X_bounds[jj][0], X_bounds[jj][1])
            else:
                test_X[ii, jj] = (
                    np.random.uniform() * (X_bounds[jj][1] - X_bounds[jj][0])
                    + X_bounds[jj][0]
                )
    return test_X


def predict(
    X: ArrayLike, y: ArrayLike, test_X: ArrayLike, nu: floating = 1.5
) -> Tuple[ArrayLike, ArrayLike]:
    gp, norm_mean, norm_stddev = fit_normalized_gaussian_process(X, y, nu=nu)
    y_pred, y_std = gp.predict([test_X], return_std=True)
    y_std_norm = y_std * norm_stddev
    y_pred_norm = (y_pred * norm_stddev) + norm_mean
    return y_pred_norm[0], y_std_norm[0]


def train_gaussian_process(
    sample_X: ArrayLike,
    sample_y: ArrayLike,
    X_bounds: Optional[ArrayLike] = None,
    current_X: ArrayLike = None,
    nu: floating = 1.5,
    max_samples: integer = 100,
) -> Tuple[sklearn_gaussian.GaussianProcessRegressor, floating, floating]:
    """Trains a Gaussian Process function from sample_X, sample_y data.

    Handles the case where there are other training runs in flight (current_X)

    Arguments:
        sample_X: vector of already evaluated sets of hyperparameters
        sample_y: vector of already evaluated loss function values
        X_bounds: minimum and maximum values for every dimension of X
        current_X: hyperparameters currently being explored
        nu: input to the Matern function, higher numbers make it smoother 0.5, 1.5, 2.5 are good values
         see http://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html

    Returns:
        gp: the gaussian process function
        y_mean: mean
        y_stddev: stddev

        To make a prediction with gp on real world data X, need to call:
        (gp.predict(X) * y_stddev) + y_mean
    """
    if current_X is not None:
        current_X = np.array(current_X)
        if len(current_X.shape) != 2:
            raise ValueError("Current X must be a 2 dimensional array")

        # we can't let the current samples be bigger than max samples
        # because we need to use some real samples to build the curve
        if current_X.shape[0] > max_samples - 5:
            print(
                "current_X is bigger than max samples - 5 so dropping some currently running parameters"
            )
            current_X = current_X[: (max_samples - 5), :]  # type: ignore
    if len(sample_y.shape) != 1:
        raise ValueError("Sample y must be a 1 dimensional array")

    if sample_X.shape[0] != sample_y.shape[0]:
        raise ValueError(
            "Sample X and sample y must be the same size {} {}".format(
                sample_X.shape[0], sample_y.shape[0]
            )
        )

    if X_bounds is not None and sample_X.shape[1] != len(X_bounds):
        raise ValueError(
            "Bounds must be the same length as Sample X's second dimension"
        )

    # gaussian process takes a long time to train, so if there's more than max_samples
    # we need to sample from it
    if sample_X.shape[0] > max_samples:
        sample_indices = np.random.randint(sample_X.shape[0], size=max_samples)
        X = sample_X[sample_indices]
        y = sample_y[sample_indices]
    else:
        X = sample_X
        y = sample_y
    gp, y_mean, y_stddev = fit_normalized_gaussian_process(X, y, nu=nu)
    if current_X is not None:
        # if we have some hyperparameters running, we pretend that they return
        # the prediction of the function we've fit
        X = np.append(X, current_X, axis=0)
        current_y_fantasy = (gp.predict(current_X) * y_stddev) + y_mean
        y = np.append(y, current_y_fantasy)
        gp, y_mean, y_stddev = fit_normalized_gaussian_process(X, y, nu=nu)
    return gp, y_mean, y_stddev


def filter_nans(sample_X: ArrayLike, sample_y: ArrayLike) -> ArrayLike:
    is_row_finite = ~(np.isnan(sample_X).any(axis=1) | np.isnan(sample_y))
    sample_X = sample_X[is_row_finite, :]
    sample_y = sample_y[is_row_finite]
    return sample_X, sample_y


def next_sample(
    *,
    sample_X: ArrayLike,
    sample_y: ArrayLike,
    X_bounds: Optional[ArrayLike] = None,
    current_X: Optional[ArrayLike] = None,
    nu: floating = 1.5,
    max_samples_for_gp: integer = 100,
    improvement: floating = 0.01,
    num_points_to_try: integer = 1000,
    opt_func: str = "expected_improvement",
    test_X: Optional[ArrayLike] = None,
) -> Tuple[ArrayLike, floating, floating, floating, floating]:
    """Calculates the best next sample to look at via bayesian optimization.

    Args:
        sample_X: ArrayLike, shape (N_runs, N_params)
            2d array of already evaluated sets of hyperparameters
        sample_y: ArrayLike, shape (N_runs,)
            1d array of already evaluated loss function values
        X_bounds: ArrayLike, optional, shape (N_params, 2), default None
            2d array minimum and maximum values for every dimension of X
        current_X: ArrayLike, optional, shape (N_runs_in_flight, N_params), default None
            hyperparameters currently being explored
        nu: floating, optional, default = 1.5
            input to the Matern function, higher numbers make it smoother. 0.5,
            1.5, 2.5 are good values  see

               http://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html

        max_samples_for_gp: integer, optional, default 100
            maximum samples to consider (since algo is O(n^3)) for performance,
            but also adds some randomness. this number of samples will be chosen
            randomly from the sample_X and used to train the GP.
        improvement: floating, optional, default 0.1
            amount of improvement to optimize for -- higher means take more exploratory risks
        num_points_to_try: integer, optional, default 1000
            number of X values to try when looking for value with highest expected probability
            of improvement
        opt_func: one of {"expected_improvement", "prob_of_improvement"} - whether to optimize expected
                improvement of probability of improvement.  Expected improvement is generally better - may want
                to remove probability of improvement at some point.  (But I think prboability of improvement
                is a little easier to calculate)
        test_X: X values to test when looking for the best values to try

    Returns:
        suggested_X: optimal X value to try
        prob_of_improvement: probability of an improvement
        predicted_y: predicted value
        predicted_std: stddev of predicted value
        expected_improvement: expected improvement
    """
    # Sanity check the data
    sample_X = np.array(sample_X)
    sample_y = np.array(sample_y)
    if test_X is not None:
        test_X = np.array(test_X)
    if len(sample_X.shape) != 2:
        raise ValueError("Sample X must be a 2 dimensional array")

    if len(sample_y.shape) != 1:
        raise ValueError("Sample y must be a 1 dimensional array")

    if sample_X.shape[0] != sample_y.shape[0]:
        raise ValueError("Sample X and y must be same length")

    if test_X is not None:
        # if test_X is set, usually this is for simulation/testing
        if X_bounds is not None:
            raise ValueError("Can't set test_X and X_bounds")

    else:
        # normal case where we randomly sample our test_X
        if X_bounds is None:
            raise ValueError("Must pass in test_X or X_bounds")

    filtered_X, filtered_y = filter_nans(sample_X, sample_y)

    # we can't run this algothim with less than two sample points, so we'll
    # just return a random point
    if filtered_X.shape[0] < 2:
        if test_X is not None:
            # pick a random row from test_X
            row = np.random.choice(test_X.shape[0])
            X = test_X[row, :]
        else:
            X = random_sample(X_bounds, 1)[0]
        if filtered_X.shape[0] < 1:
            prediction = 0.0
        else:
            prediction = filtered_y[0]
        return (
            X,
            1.0,
            prediction,
            np.nan,
            np.nan,
        )

    # build the acquisition function
    gp, y_mean, y_stddev, = train_gaussian_process(
        filtered_X, filtered_y, X_bounds, current_X, nu, max_samples_for_gp
    )
    # Look for the minimum value of our fitted-target-function + (kappa * fitted-target-std_dev)
    if test_X is None:  # this is the usual case
        test_X = random_sample(X_bounds, num_points_to_try)
    y_pred, y_pred_std = gp.predict(test_X, return_std=True)

    # best value of y we've seen so far.  i.e. y*
    min_unnorm_y = np.min(filtered_y)
    # hack for dealing with predicted std of 0
    epsilon = 0.00000001

    """
    if opt_func == "probability_of_improvement":
        min_norm_y = (min_unnorm_y - y_mean) / y_stddev - improvement
    else:
    """
    min_norm_y = (min_unnorm_y - y_mean) / y_stddev

    Z = -(y_pred - min_norm_y) / (y_pred_std + epsilon)
    prob_of_improve: np.ndarray = scipy_stats.norm.cdf(Z)
    e_i = -(y_pred - min_norm_y) * scipy_stats.norm.cdf(
        Z
    ) + y_pred_std * scipy_stats.norm.pdf(Z)

    """
    if opt_func == "probability_of_improvement":
        best_test_X_index = np.argmax(prob_of_improve)
    else:
    """
    best_test_X_index = np.argmax(e_i)

    suggested_X = test_X[best_test_X_index]
    suggested_X_prob_of_improvement = prob_of_improve[best_test_X_index]
    suggested_X_predicted_y = y_pred[best_test_X_index] * y_stddev + y_mean
    suggested_X_predicted_std = y_pred_std[best_test_X_index] * y_stddev

    # recalculate expected improvement
    min_norm_y = (min_unnorm_y - y_mean) / y_stddev
    z_best = -(y_pred[best_test_X_index] - min_norm_y) / (
        y_pred_std[best_test_X_index] + epsilon
    )
    suggested_X_expected_improvement = -(
        y_pred[best_test_X_index] - min_norm_y
    ) * scipy_stats.norm.cdf(z_best) + y_pred_std[
        best_test_X_index
    ] * scipy_stats.norm.pdf(
        z_best
    )

    return (
        suggested_X,
        suggested_X_prob_of_improvement,
        suggested_X_predicted_y,
        suggested_X_predicted_std,
        suggested_X_expected_improvement,
    )


def bayes_search_next_run(
    runs: List[SweepRun],
    config: Union[dict, SweepConfig],
    validate: bool = False,
    minimum_improvement: floating = 0.1,
) -> SweepRun:
    """Suggest runs using Bayesian optimization.

    >>> suggestion = bayes_search_next_run([], {
    ...    'method': 'bayes',
    ...    'parameters': {'a': {'min': 1., 'max': 2.}},
    ...    'metric': {'name': 'loss', 'goal': 'maximize'}
    ... })

    Args:
        runs: The runs in the sweep.
        config: The sweep's config.
        minimum_improvement: The minimium improvement to optimize for. Higher means take more exploratory risks.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.

    Returns:
        The suggested run.
    """

    if validate:
        config = SweepConfig(config)

    if "metric" not in config:
        raise ValueError('Bayesian search requires "metric" section')

    if config["method"] != "bayes":
        raise ValueError("Invalid sweep configuration for bayes_search_next_run.")

    goal = config["metric"]["goal"]
    metric_name = config["metric"]["name"]
    worst_func = min if goal == "maximize" else max
    params = HyperParameterSet.from_config(config["parameters"])

    if len(params.searchable_params) == 0:
        raise ValueError("Need at least one searchable parameter for bayes search.")

    sample_X: ArrayLike = []
    current_X: ArrayLike = []
    y: ArrayLike = []

    X_bounds = [[0.0, 1.0]] * len(params.searchable_params)

    # we calc the max metric to put as the metric for failed runs
    # so that our bayesian search stays away from them
    worst_metric: floating = 0.0
    for run in runs:
        if run.state == RunState.finished:
            try:
                run_extremum = run.metric_extremum(
                    metric_name, kind="minimum" if goal == "maximize" else "maximum"
                )
            except ValueError:
                run_extremum = 0.0  # default
            worst_metric = worst_func(worst_metric, run_extremum)

    X_norms = params.convert_runs_to_normalized_vector(runs)
    for run, X_norm in zip(runs, X_norms):
        if run.state == RunState.finished:
            try:
                metric = run.metric_extremum(
                    metric_name, kind="maximum" if goal == "maximize" else "minimum"
                )
            except ValueError:
                metric = 0.0  # default
            y.append(metric)
            sample_X.append(X_norm)
        elif run.state in [RunState.running, RunState.preempting, RunState.preempted]:
            # run is in progress
            # we wont use the metric, but we should pass it into our optimizer to
            # account for the fact that it is running
            current_X.append(X_norm)
        elif run.state in [RunState.failed, RunState.crashed, RunState.killed]:
            # run failed, but we're still going to use it
            # maybe we should be smarter about this
            y.append(worst_metric)
            sample_X.append(X_norm)
        else:
            raise ValueError("Run is in unknown state")

    if len(sample_X) == 0:
        sample_X = np.empty([0, 0])
    else:
        sample_X = np.asarray(sample_X)

    if len(current_X) > 0:
        current_X = np.array(current_X)

    # impute bad metric values from y
    y = np.asarray(y)
    if len(y) > 0:
        y[~np.isfinite(y)] = worst_metric

    # next_sample is a minimizer, so if we are trying to
    # maximize, we need to negate y
    y *= -1 if goal == "maximize" else 1

    (
        suggested_X,
        suggested_X_prob_of_improvement,
        suggested_X_predicted_y,
        suggested_X_predicted_std,
        suggested_X_expected_improvement,
    ) = next_sample(
        sample_X=sample_X,
        sample_y=y,
        X_bounds=X_bounds,
        current_X=current_X if len(current_X) > 0 else None,
        improvement=minimum_improvement,
    )

    # convert the parameters from vector of [0,1] values
    # to the original ranges
    for param in params:
        if param.type == HyperParameter.CONSTANT:
            continue
        try_value = suggested_X[params.param_names_to_index[param.name]]
        param.value = param.ppf(try_value)

    ret_dict = params.to_config()
    info = {
        "success_probability": suggested_X_prob_of_improvement,
        "predicted_value": suggested_X_predicted_y,
        "predicted_value_std_dev": suggested_X_predicted_std,
        "expected_improvement": suggested_X_expected_improvement,
    }
    return SweepRun(config=ret_dict, search_info=info)



import pytest
from .. import SweepRun, RunState


def test_skipped_steps():
    run = SweepRun(
        name="a",
        state=RunState.running,
        history=[
            {"loss": 10},
            {"a": 9},
            {"a": 8},
            {"a": 7},
            {"loss": 6},
            {"a": 5},
            {"a": 4},
            {"a": 3},
            {"a": 2},
            {"loss": 1},
        ],
    )
    assert run.metric_history("loss") == [10, 6, 1]


def test_summary_metric_none():
    run = SweepRun(
        name="a",
        summary_metrics=None,
    )
    with pytest.raises(ValueError):
        run.summary_metric("loss")



from typing import List, Union, Dict, Any

import numpy as np

from .config import SweepConfig
from .run import SweepRun, RunState


def hyperband_stop_runs(
    runs: List[SweepRun],
    config: Union[dict, SweepConfig],
    validate: bool = False,
) -> List[SweepRun]:
    """
    Suggest sweep runs to terminate early using Hyperband: A Novel Bandit-Based Approach to Hyperparameter Optimization
      https://arxiv.org/pdf/1603.06560.pdf


    >>> to_stop = hyperband_stop_runs(
    ...    [SweepRun(
    ...        name="a",
    ...        state=RunState.finished,  # This is already stopped
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 9},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="b",
    ...        state=RunState.running,  # This should be stopped
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 10},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="c",
    ...        state=RunState.running,  # This passes band 1 but not band 2
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 8},
    ...            {"loss": 8},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="d",
    ...        state=RunState.running,
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 7},
    ...            {"loss": 7},
    ...        ],
    ...    ),
    ...    SweepRun(
    ...        name="e",
    ...        state=RunState.finished,
    ...        history=[
    ...            {"loss": 10},
    ...            {"loss": 6},
    ...            {"loss": 6},
    ...        ],
    ...    ),
    ... ],
    ... {
    ...    "method": "grid",
    ...    "metric": {"name": "loss", "goal": "minimize"},
    ...    "early_terminate": {
    ...        "type": "hyperband",
    ...        "max_iter": 5,
    ...        "eta": 2,
    ...        "s": 2,
    ...    },
    ...    "parameters": {"a": {"values": [1, 2, 3]}},
    ... })

    Args:
        runs: The runs in the sweep.
        config: The sweep's config.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.

    Returns:
        List of runs to stop early.
    """

    # validate config and fill in defaults
    if validate:
        config = SweepConfig(config)

    if "metric" not in config:
        raise ValueError('Hyperband stopping requires "metric" section')

    if "early_terminate" not in config:
        raise ValueError('Hyperband stopping requires "early_terminate" section.')
    et_config = config["early_terminate"]

    if et_config["type"] != "hyperband":
        raise ValueError("Sweep config is not configured for hyperband stopping")

    if "max_iter" in et_config:
        max_iter = et_config["max_iter"]
        s = et_config["s"]
        eta = et_config["eta"]

        band = max_iter
        bands = []
        for i in range(s):
            band /= eta
            if band < 1:
                break
            bands.append(int(band))
        bands = sorted(bands)

    # another way of defining hyperband with min_iter and possibly eta
    elif "min_iter" in et_config:
        min_iter = et_config["min_iter"]
        eta = et_config["eta"]

        band = min_iter
        bands = []
        for i in range(100):
            bands.append(int(band))
            band *= eta
    else:
        raise ValueError(
            'invalid config for hyperband stopping: either "max_iter" or "min_iter" must be specified'
        )
    r = 1.0 / eta

    if len(bands) < 1:
        raise ValueError("Bands must be an array of length at least 1")
    if r < 0 or r > 1:
        raise ValueError("r must be a float between 0 and 1")

    terminate_runs: List[SweepRun] = []
    metric_name = config["metric"]["name"]

    all_run_histories = []  # we're going to look at every run
    for run in runs:
        history = run.metric_history(metric_name)
        if config["metric"]["goal"] == "maximize":
            history = list(map(lambda x: -x, history))
        if len(history) > 0:
            all_run_histories.append(history)

    thresholds = []
    # iterate over the histories at every band and find the threshold for a run to be in the top r percentile
    for band in bands:
        # values of metric at iteration number "band"
        band_values = [h[band] for h in all_run_histories if len(h) > band]
        if len(band_values) == 0:
            threshold = np.inf
        else:
            threshold = sorted(band_values)[int((r) * len(band_values))]
        thresholds.append(threshold)

    info: Dict[str, Any] = {}
    info["lines"] = []
    info["lines"].append(
        "Bands: %s"
        % (
            ", ".join(
                [
                    "%s = %s" % (band, threshold)
                    for band, threshold in zip(bands, thresholds)
                ]
            )
        )
    )

    info["bands"] = bands
    info["thresholds"] = thresholds

    for run in runs:
        if run.state == RunState.running:
            history = run.metric_history(metric_name)
            if config["metric"]["goal"] == "maximize":
                history = list(map(lambda x: -x, history))

            closest_band = -1
            closest_threshold = 0.0
            bandstr = ""
            termstr = ""
            for band, threshold in zip(bands, thresholds):
                if band < len(history):
                    closest_band = band
                    closest_threshold = threshold
                else:
                    break

            if closest_band != -1:  # no bands apply yet
                bandstr = " (Metric: %f Band: %d Threshold %f)" % (
                    min(history),
                    closest_band,
                    closest_threshold,
                )
                if min(history) > closest_threshold:
                    terminate_runs.append(run)
                    termstr = " STOP"

            run_info = info.copy()
            run_info["lines"].append(
                "Run: %s Step: %d%s%s" % (run.name, len(history), bandstr, termstr)
            )
            run.early_terminate_info = run_info

    return terminate_runs



from .config.cfg import SweepConfig
from .run import SweepRun
from .params import HyperParameterSet

from typing import Union


def random_search_next_run(
    sweep_config: Union[dict, SweepConfig], validate: bool = False
) -> SweepRun:
    """Suggest runs with Hyperparameters sampled randomly from specified distributions.

    >>> suggestion = random_search_next_run({'method': 'random', 'parameters': {'a': {'min': 1., 'max': 2.}}})

    Args:
        sweep_config: The sweep's config.
        validate: Whether to validate `sweep_config` against the SweepConfig JSONschema.
           If true, will raise a Validation error if `sweep_config` does not conform to
           the schema. If false, will attempt to run the sweep with an unvalidated schema.

    Returns:
        The suggested run.
    """

    # ensure that the sweepconfig is properly formatted
    if validate:
        sweep_config = SweepConfig(sweep_config)

    if sweep_config["method"] != "random":
        raise ValueError("Invalid sweep configuration for random_search_next_run.")

    params = HyperParameterSet.from_config(sweep_config["parameters"])

    for param in params:
        param.value = param.sample()

    return SweepRun(config=params.to_config())



from .. import stop_runs, next_run, RunState, SweepRun


def test_hyperband_min_iter_bands():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "min_iter": 3,
            "eta": 3,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"][:3] == [3, 9, 27]


def test_hyperband_min_iter_bands_max():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "accuracy", "goal": "maximize"},
        "early_terminate": {
            "type": "hyperband",
            "min_iter": 3,
            "eta": 3,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"accuracy": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"accuracy": 10 + i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"][:3] == [3, 9, 27]


def test_hyperband_max_iter_bands():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 81,
            "eta": 3,
            "s": 3,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"][:3] == [3, 9, 27]


def test_init_from_max_iter():
    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 18,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10} for _ in range(4)]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])

    assert to_stop[0] is run
    assert to_stop[0].early_terminate_info["bands"] == [2, 6]


def test_single_run():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 18,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run])
    assert len(to_stop) == 0


def test_2runs_band1_pass():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 18,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    run = next_run(sweep_config, [])
    run.state = RunState.running
    run.history = [{"loss": 10}, {"loss": 10}, {"loss": 6}]
    run2 = next_run(sweep_config, [run])
    run2.state = RunState.running
    run2.history = [{"loss": 10 - i} for i in range(10)]
    to_stop = stop_runs(sweep_config, [run, run2])
    assert len(to_stop) == 0


def test_5runs_band1_stop_2():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 5,
            "eta": 2,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": 10},
                {"loss": 9},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": 10},
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This passes band 1 but not band 2
            history=[
                {"loss": 10},
                {"loss": 8},
                {"loss": 8},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": 10},
                {"loss": 7},
                {"loss": 7},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.finished,
            history=[
                {"loss": 10},
                {"loss": 6},
                {"loss": 6},
            ],
        ),
    ]

    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == runs[1:3]


def test_5runs_band1_stop_2_1stnoband():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 5,
            "eta": 2,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": 10},
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This passes band 1 but not band 2
            history=[
                {"loss": 10},
                {"loss": 8},
                {"loss": 8},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": 10},
                {"loss": 7},
                {"loss": 7},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.finished,
            history=[
                {"loss": 10},
                {"loss": 6},
                {"loss": 6},
            ],
        ),
    ]

    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == runs[1:3]


def test_eta_3():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "minimize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 9,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": 10},
                {"loss": 9},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": 10},
                {"loss": 10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This fails the first threeshold but snuck in so we wont kill
            history=[
                {"loss": 10},
                {"loss": 8},
                {"loss": 8},
                {"loss": 3},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": 10},
                {"loss": 7},
                {"loss": 7},
                {"loss": 4},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.running,  # this passes band 1 but doesn't pass band 2
            history=[
                {"loss": 10},
                {"loss": 6},
                {"loss": 6},
                {"loss": 6},
            ],
        ),
    ]

    # bands are at 1 and 3, thresholds are 7 and 4
    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == [runs[1], runs[-1]]


def test_eta_3_max():

    sweep_config = {
        "method": "grid",
        "metric": {"name": "loss", "goal": "maximize"},
        "early_terminate": {
            "type": "hyperband",
            "max_iter": 9,
            "eta": 3,
            "s": 2,
        },
        "parameters": {"a": {"values": [1, 2, 3]}},
    }

    runs = [
        SweepRun(
            name="a",
            state=RunState.finished,  # This wont be stopped because already stopped
            history=[
                {"loss": -10},
                {"loss": -9},
            ],
        ),
        SweepRun(
            name="b",
            state=RunState.running,  # This should be stopped
            history=[
                {"loss": -10},
                {"loss": -10},
            ],
        ),
        SweepRun(
            name="c",
            state=RunState.running,  # This fails the first threeshold but snuck in so we wont kill
            history=[
                {"loss": -10},
                {"loss": -8},
                {"loss": -8},
                {"loss": -3},
            ],
        ),
        SweepRun(
            name="d",
            state=RunState.running,
            history=[
                {"loss": -10},
                {"loss": -7},
                {"loss": -7},
                {"loss": -4},
            ],
        ),
        SweepRun(
            name="e",
            state=RunState.running,  # this passes band 1 but doesn't pass band 2
            history=[
                {"loss": -10},
                {"loss": -6},
                {"loss": -6},
                {"loss": -6},
            ],
        ),
    ]

    # bands are at 1 and 3, thresholds are 7 and 4
    to_stop = stop_runs(sweep_config, runs)
    assert to_stop == [runs[1], runs[-1]]



import json
import jsonref
import jsonschema
from jsonschema import Draft7Validator, validators

from pathlib import Path
from typing import Dict, Optional, Tuple
from copy import deepcopy

sweep_config_jsonschema_fname = Path(__file__).parent / "schema.json"
with open(sweep_config_jsonschema_fname, "r") as f:
    sweep_config_jsonschema = json.load(f)


dereferenced_sweep_config_jsonschema = jsonref.JsonRef.replace_refs(
    sweep_config_jsonschema
)

format_checker = jsonschema.FormatChecker()


@format_checker.checks("float")
def float_checker(value):
    return isinstance(value, float)


@format_checker.checks("integer")
def int_checker(value):
    return isinstance(value, int)


validator = Draft7Validator(
    schema=sweep_config_jsonschema, format_checker=format_checker
)


def extend_with_default(validator_class):
    # https://python-jsonschema.readthedocs.io/en/stable/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):

        errored = False
        for error in validate_properties(
            validator,
            properties,
            instance,
            schema,
        ):
            errored = True
            yield error

        if not errored:
            for property, subschema in properties.items():
                if "default" in subschema:
                    instance.setdefault(property, subschema["default"])

    return validators.extend(
        validator_class,
        {"properties": set_defaults},
    )


DefaultFiller = extend_with_default(Draft7Validator)
default_filler = DefaultFiller(
    schema=sweep_config_jsonschema, format_checker=format_checker
)


def fill_parameter(config: Dict) -> Optional[Tuple[str, Dict]]:
    # names of the parameter definitions that are allowed
    allowed_schemas = [
        d["$ref"].split("/")[-1]
        for d in sweep_config_jsonschema["definitions"]["parameter"]["anyOf"]
    ]

    for schema_name in allowed_schemas:
        # create a jsonschema object to validate against the subschema
        subschema = dereferenced_sweep_config_jsonschema["definitions"][schema_name]

        try:
            jsonschema.Draft7Validator(
                subschema, format_checker=format_checker
            ).validate(config)
        except jsonschema.ValidationError:
            continue
        else:
            filler = DefaultFiller(subschema, format_checker=format_checker)

            # this sets the defaults, modifying config inplace
            config = deepcopy(config)
            filler.validate(config)
            return schema_name, config

    return None


def fill_schema(d: Dict) -> Dict:
    from . import schema_violations_from_proposed_config

    # check that the schema is valid
    violations = schema_violations_from_proposed_config(d)
    if len(violations) != 0:
        raise jsonschema.ValidationError("\n".join(violations))

    validated = deepcopy(d)

    # update the parameters
    filled = {}
    for k, v in validated["parameters"].items():
        result = fill_parameter(v)
        if result is None:
            raise jsonschema.ValidationError(f"Parameter {k} is malformed")
        _, config = result
        filled[k] = config
    validated["parameters"] = filled

    if "early_terminate" in validated:
        if validated["early_terminate"]["type"] == "hyperband":
            filler = DefaultFiller(
                schema=dereferenced_sweep_config_jsonschema["definitions"][
                    "hyperband_stopping"
                ],
                format_checker=format_checker,
            )
            filler.validate(validated["early_terminate"])

    return validated



import pytest
import jsonschema
from .. import config


def test_nvalid_sweep_config_nonuniform_array_elements_categorical():
    invalid_config = {
        "method": "grid",
        "parameters": {
            "v1": {"values": [None, 2, 3, "a", (2, 3), 3]},
        },
    }

    with pytest.raises(jsonschema.ValidationError):
        _ = config.SweepConfig(invalid_config)

    invalid_config = {
        "method": "grid",
        "parameters": {
            "v1": {"values": [None, 2, 3, "a", (2, 3)]},
        },
    }

    # doesn't raise
    _ = config.SweepConfig(invalid_config)


def test_min_max_validation():
    invalid_config = {
        "method": "random",
        "parameters": {
            "v1": {"max": 3, "min": 5},
            "v2": {"min": 5, "max": 6},
        },
    }

    with pytest.raises(jsonschema.ValidationError):
        _ = config.SweepConfig(invalid_config)


def test_negative_sigma_validation():
    invalid_config = {
        "method": "random",
        "parameters": {
            "v1": {"mu": 0.1, "sigma": -0.1, "distribution": "normal"},
        },
    }
    with pytest.raises(jsonschema.ValidationError):
        _ = config.SweepConfig(invalid_config)



#
"""Feature Flags Module.

This module implements a feature flag system for the wandb library to require experimental features
and notify the user when features have been deprecated.

Example:
    import wandb
    wandb.require("wandb-service@beta")
    wandb.require("incremental-artifacts@beta")
"""

from typing import Sequence, Union

import wandb
from wandb.errors import RequireError


class _Requires(object):
    """Internal feature class."""

    _features: Sequence[str]

    def __init__(self, features: Union[str, Sequence[str]]) -> None:
        self._features = (
            tuple([features]) if isinstance(features, str) else tuple(features)
        )

    def require_require(self) -> None:
        pass

    def apply(self) -> None:
        """Call require_* method for supported features."""
        last_message: str = ""
        for feature_item in self._features:
            full_feature = feature_item.split("@", 2)[0]
            feature = full_feature.split(":", 2)[0]
            func_str = "require_{}".format(feature.replace("-", "_"))
            func = getattr(self, func_str, None)
            if not func:
                last_message = "require() unsupported requirement: {}".format(feature)
                wandb.termwarn(last_message)
                continue
            func()

        if last_message:
            wandb.termerror(
                "Supported wandb.require() features can be found at: http://wandb.me/library-require"
            )
            raise RequireError(last_message)


def require(
    requirement: Union[str, Sequence[str]] = None,
    experiment: Union[str, Sequence[str]] = None,
) -> None:
    """Indicate which experimental features are used by the script.

    Args:
        requirement: (str or list) Features to require
        experiment: (str or list) Features to require

    Raises:
        wandb.errors.RequireError: if not supported or other error
    """
    features = requirement or experiment
    if not features:
        return

    f = _Requires(features=features)
    f.apply()
